{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeOperators #-}

module Test.Core.LinearAlgebra where

import Data.Bifunctor (first)
import Data.Bimap (Bimap)
import Data.Bimap qualified as Bimap
import Data.Map.Strict (Map)
import Data.Map.Strict qualified as M
import Gargantext.Core.LinearAlgebra qualified as LA
import Gargantext.Core.Methods.Matrix.Accelerate.Utils qualified as Legacy
import Gargantext.Core.Methods.Similarities.Accelerate.Distributional qualified as Legacy
import Gargantext.Core.Viz.Graph.Index qualified as Legacy
import Gargantext.Orphans.Accelerate ()
import Prelude
import Test.Tasty
import Test.Tasty.QuickCheck
import Data.Array.Accelerate hiding (Ord, Eq, map)
import Data.Array.Accelerate.Interpreter qualified as Naive
import Data.Array.Accelerate.LLVM.Native qualified as LLVM

compareImplementations :: (Arbitrary a, Eq b, Show b)
                       => (a -> b)
                       -> (a -> c)
                       -> (c -> b)
                       -> a
                       -> Property
compareImplementations implementation1 implementation2 mapResults inputData
  = implementation1 inputData === mapResults (implementation2 inputData)

compareTermDivNan :: (Array TermDivNanShape Double)
                  -> (Array TermDivNanShape Double)
                  -> Property
compareTermDivNan i1 i2
  = Naive.run (Legacy.termDivNan (use i1) (use i2)) === Naive.run (Legacy.termDivNan (use i1) (use i2))

compareDistributional :: Matrix Int
                      -> Property
compareDistributional i1
  = Legacy.distributionalWith Naive.run i1 === Legacy.distributionalWith LLVM.run i1

mapCreateIndices :: Ord t => (Map t Legacy.Index, Map Legacy.Index t) -> Bimap LA.Index t
mapCreateIndices (_m1, m2) = Bimap.fromList $ map (first LA.Index) $ M.toList m2

type TermDivNanShape = Z :. Int :. Int

twoByTwo :: Matrix Int
twoByTwo = fromList (Z :. 2 :. 2) (Prelude.replicate 4 0)

tests :: TestTree
tests = testGroup "LinearAlgebra" [
      testProperty "createIndices roundtrip" (compareImplementations (LA.createIndices @Int @Int) Legacy.createIndices mapCreateIndices)
    , testProperty "termDivNan"     compareTermDivNan
    , testGroup "distributional" [
        testProperty "2x2" (compareDistributional twoByTwo)
      , testProperty "3x2" (compareDistributional $ fromList (Z :. 3 :. 2) (Prelude.replicate 6 0))
      , testProperty "roundtrips" (compareImplementations (Legacy.distributionalWith Naive.run)
                                                              (Legacy.distributionalWith LLVM.run)
                                                              id)
      ]
    ]
